# Made by Marcos Boggia
from gui_automation.gui_automation import GuiAuto
from gui_automation.image_detector import TMDetector, MultiscaledTMDetector, SQDIFF, CCOEFF, CCORR
from gui_automation.foreground_handler import ForegroundHandler
from gui_automation.background_handler import BackgroundHandlerWin32
from gui_automation.image_loader import load_images
name = "gui_automation"
